// Copyright (C) 2001,2002,2003 Krzysztof Drewicz <hunter@pld.org.pl>
// 
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 1, or (at your option)
// any later version. See <URL:http://www.gnu.org/copyleft/gpl.html>
// for a copy.
// 

// Simple daemon for mks32 (http://www.mks.com.pl)
// see http://www.raszyn.pl/~hunter/howto/mks.html 
//     http://www.raszyn.pl/~hunter/howto/demon.html for more info.

#include <stdio.h>
#include <sys/times.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <errno.h>
#include <netinet/in.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <syslog.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/resource.h>

#include "mks.h"

void err(char *msg)
{
    syslog(LOG_ERR,"%s",msg);
    perror(msg);
    perror(strerror(errno));
    exit(1);
}

#define ILEINITSTRINGS	4
char mks_version[INITBUFF];
char mks_bversion[INITBUFF];

int init_mks(FILE * m_in, FILE * m_out, FILE * m_err)
{
    char t[MAXBUFF];
    char co[ILEINITSTRINGS][INITBUFF] =
	{ "init...\0", "mks_vir: database version \0", "init OK\0", "check file(s) from stdin list\0" };

    int i;
#ifdef  DEBUG_INIT
	struct timeval ru,ru2;
	struct timezone tzp;

 if ((i=gettimeofday(&ru2,&tzp))!=0) err("init_mks:getrusage");

#endif

    memset(t, 0, MAXBUFF);
    memset(mks_version, 0, INITBUFF);
    for (i = 0; i < ILEINITSTRINGS; i++) {
	fgets(t, MAXBUFF - 1, m_err);
	if (strstr(t, co[i]) == NULL) {
		syslog(LOG_WARNING,"error init: %d %s ", i, t );
		return -1;
		}
    if (i==0) strncpy(mks_version,t,INITBUFF-1);
    if (i==1) strncpy(mks_bversion,t,INITBUFF-1);
    }

#ifdef  DEBUG_INIT
 if ((i=gettimeofday(&ru,&tzp))!=0) err("init_mks:getrusage");
	syslog(LOG_NOTICE,"init took: %2.3f s \n", (float)(ru.tv_sec-ru2.tv_sec+((float)(ru.tv_usec-ru2.tv_usec))/1000000));
#endif
	return 0;
}

void close_all_fds()
{
    int i;
    int max_fd;

    max_fd = getdtablesize();
    for (i = 0; i < max_fd; i++)
		close(i);
}


// obsuga sygnaw:
// pid procesu mks32
int exit_flag=0;

int reload_flag=0;

void on_usr1()
{
		reload_flag=1;
}

void on_signal()
{
    syslog(LOG_NOTICE,"SIGTERM received, terminating");
	exit_flag=1;
}

void on_signal_synek(int fd)
{
 shutdown(fd,SHUT_RDWR);
 syslog(LOG_NOTICE,"SIGTERM received while scanning, terminating");
}


void on_mks_exit()
{
    int i;
    int max_fd;

    signal(SIGCHLD, SIG_IGN);
    perror("child process mks32 died");
    perror(strerror(errno));
    close_all_fds();
    exit(3);
}

// ilu jest aktualnie klientw
int child_count = 0;
long child_served = 0;

void statystyki()
{
	struct rusage ru,ru1;
    int i;

    if ((i=getrusage(RUSAGE_SELF,&ru))!=0) err("stats:getrusage");
    if ((i=getrusage(RUSAGE_CHILDREN,&ru1))!=0) err("stats:getrusage");

	syslog (LOG_INFO, "ch s/n :%d/%ld t u/s: %3.2f/%3.2f ch u/s: %5.2f/%5.2f", 
			child_served, child_count, 
            (float) (ru.ru_utime.tv_sec+(float)ru.ru_utime.tv_usec/1000000), 
            (float) (ru.ru_stime.tv_sec+(float)ru.ru_stime.tv_usec/1000000), 
            (float) (ru1.ru_utime.tv_sec+(float)ru1.ru_utime.tv_usec/1000000), 
            (float) (ru1.ru_stime.tv_sec+(float)ru1.ru_stime.tv_usec/1000000)); 
    signal(SIGALRM, (sig_t) statystyki);
	signal(SIGUSR1, (sig_t) statystyki);
	alarm(DEBUG_STATS_TIME);
}

void statystyki_file(FILE *f) 
{        
struct rusage ru,ru1;
int i;

    if ((i=getrusage(RUSAGE_SELF,&ru))!=0) err("stats:getrusage");
    if ((i=getrusage(RUSAGE_CHILDREN,&ru1))!=0) err("stats:getrusage");

    fprintf (f, "ch s/n :%d/%ld t u/s: %3.2f/%3.2f ch u/s: %5.2f/%5.2f\n",
        child_served, child_count,
        (float) (ru.ru_utime.tv_sec+(float)ru.ru_utime.tv_usec/1000000),
        (float) (ru.ru_stime.tv_sec+(float)ru.ru_stime.tv_usec/1000000),
        (float) (ru1.ru_utime.tv_sec+(float)ru1.ru_utime.tv_usec/1000000),
        (float) (ru1.ru_stime.tv_sec+(float)ru1.ru_stime.tv_usec/1000000));
}

void on_child_exit()
{
        extern child_count;
        extern long child_served;

	while (waitpid(-1, NULL,WNOHANG)>0) {
			child_count--;
			child_served++;
	}
	signal(SIGCHLD, (sig_t) on_child_exit);
}

void on_child_timeout()
{
	syslog ( LOG_NOTICE, "Client timeout");
	exit(8);
}

void
end_mks(int fa[2], int fb[2], int fc[2], FILE * m_in, FILE * m_out,
	FILE * m_err)
{
    // wypychamy bufory
    fflush(m_in);
    fflush(m_out);
    fflush(m_err);
    // zamykamy kanay do rozmowy z mks
    fclose(m_in);
    fclose(m_out);
    fclose(m_err);
    // zamykamy ich pary z pipew
    close(fa[0]);
    close(fb[1]);
    close(fc[1]);
}

void
zrob_rury(int *a, int *b, int *c, int fa[2], int fb[2], int fc[2],
	  FILE ** min, FILE ** mout, FILE ** merr)
{
    (*a) = pipe(fa);
    (*b) = pipe(fb);
    (*c) = pipe(fc);

    if (((*a) < 0) || ((*b) < 0) || ((*c) < 0)) {
	perror("There is no pipe !");
	perror(strerror(errno));
    }
    // Przechodzimy na File* z int fd
    // fd[1] tu piszemy 
    // fd[0] ztad sie czyta

    if (((*min) = fdopen(fa[1], "w")) == NULL)
	err("mks_in");
    if (((*mout) = fdopen(fb[0], "r")) == NULL)
	err("mks_out");
    if (((*merr) = fdopen(fc[0], "r")) == NULL)
	err("mks_err");
}


ignoruj()
{
    // jestemy ignorantami:
    *signal(SIGHUP, SIG_IGN);
    signal(SIGINT, SIG_IGN);
    signal(SIGQUIT, SIG_IGN);
    signal(SIGILL, SIG_IGN);
    signal(SIGABRT, SIG_IGN);
    signal(SIGFPE, SIG_IGN);
    signal(SIGFPE, SIG_IGN);
    signal(SIGUSR1, (sig_t) statystyki );
    signal(SIGUSR2, (sig_t) on_usr1);
    signal(SIGCONT, SIG_IGN);
    signal(SIGSTOP, SIG_IGN);
    signal(SIGTTIN, SIG_IGN);
    signal(SIGTTOU, SIG_IGN);
    signal(SIGTERM, (sig_t) on_signal);	// sig_t is glibc specific
}

void run_mks(int in, int out, int error)
{
    char s[MAXBUFF];

    // Zamykamy stdin,stdout,stderr
    close(0);  close(1);   close(2);
   	// teraz zastpujemy je deskryptorami z rur
    dup2(in, 0); dup2(out, 1); dup2(error, 2);

    sprintf(s, "-f %s", CFG_FILE);
    printf("%s\n", s);
    execl(EXE_FILE, "-c", "-S", (char *) 0);
    // jeli udao si wykona execl to tutaj _nigdy_ ju nie dojdziemy
    err("OOPS: nie moge odpalic mks32");
}

void init_socket(int *sfd)
{

    struct sockaddr_in serv_addr;

    if (((*sfd) = socket(AF_INET, SOCK_STREAM, 0)) < 0)
        	err("Can't open socket");
    bzero((char *) &serv_addr, sizeof(serv_addr));
    serv_addr.sin_family = AF_INET;
    serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
    serv_addr.sin_port = htons(SERV_PORT);
    if (bind((*sfd), (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
        	err("Can't bind on socket");
    listen((*sfd), 10);
	fcntl((*sfd),O_NONBLOCK);
}

odepnij_tty()
{
	int fd;
	fd = open("/dev/tty", O_RDWR);
	ioctl(fd, TIOCNOTTY, 0);
	close(fd);
}

void init_logs()
{
		openlog("mksd",LOG_PID, LOG_MAIL);
		syslog(LOG_NOTICE, "starting");
}

lock_mks()
{
	int  fd,counter;
	char tmpname[TMP_NAME_LEN]="MKLXXXXXX\0";
	
	if (fd= mkstemp(tmpname)<0)
			err("Can't create temp file");
	while ( link (tmpname, MKS_LOCK_FILE) < 0 ) {
			if ((errno != EEXIST) || counter > 100)
					err("Link error for mks_lock");
			usleep(100);
			counter++;
	}
	
	unlink(tmpname);
}

unlock_mks()
{
		if (unlink(MKS_LOCK_FILE)<0)
				err("unlink error for mks_lock");
}


void pobierz_dane(fd,f_czytany,len)
// czyta z socketu fd dugo pliku do skanowania oraz '\n' 		
	int fd;
	FILE** f_czytany;
	int*	len;
	
{ 	
	char	s[MAXBUFF];

	(*f_czytany) = fdopen(fd, "w+");

	memset(s, MAXBUFF, 0);
	fgets(s, MAXBUFF - 1, (*f_czytany));
	(*len) = strtol(s, (char **) NULL, 10);

}

void zrob_chwilowy(f_chwilowy,f) 

	FILE** f_chwilowy;
	char	f[MAXBUFF];

{	
	int tmp_fd;

	// Generujemy nazw tymczasow:
	sprintf(f, "%s\0", "MKSXXXXXX");
	tmp_fd = mkstemp(f);
	(*f_chwilowy) = fdopen(tmp_fd, "w+");
}

kopiuj_plik_FILE (int len, FILE *f_skad,FILE *f_gdzie)
// zczytuje plik o dugoci len z f_skad i zapisuje do f_gdzie
// (fgetc i fputc s bufforowane wic nigdy nie czyta si po jednym znaku ;>
{
		while (len > 0) {
				fputc(fgetc(f_skad), f_gdzie);
				len--;
		};
}


int skanuj (mks_in, mks_out, mks_err, f, m_fifo)

	FILE *mks_in;
	FILE *mks_out;
	FILE *mks_err;
	char f[MAXBUFF];
	FILE* m_fifo;
{
		char 	s[MAXBUFF];

		fgetc(m_fifo);

		lock_mks(); // mamy wyczno na pisanie/czytanie do/z mks32

		fprintf(mks_in, "%s\n", f); fflush(mks_in);
		fgets(s, MAXBUFF - 1, mks_err); fprintf(m_fifo, "%s", s); 
#if defined(DEBUG)
       syslog(LOG_INFO, "%s", s);
#endif       
		fgets(s, MAXBUFF - 1, mks_err); fprintf(m_fifo, "%s", s); 
#if defined(DEBUG)
        syslog(LOG_INFO, "%s", s);
#endif

#if defined(OLDMKS32)
#warn old mks interface used.
#endif        
#if defined(NEWMKS)
        fgets(s, MAXBUFF - 1, mks_out); fprintf(m_fifo, "%s", s);

#if defined(DEBUG)
        syslog(LOG_INFO, "mks_OUT: %s", s);
#endif       
        
        if JEST_WIRUS_D(s) fgets(s, MAXBUFF - 1, mks_err); fprintf(m_fifo, "%s", s); 
        
#if defined(DEBUG)
        syslog(LOG_INFO, "%s", s);
#endif  
        
		fgets(s, MAXBUFF - 1, mks_err); fprintf(m_fifo, "%s", s); 
        
#if defined(DEBUG)
        syslog(LOG_INFO, "%s", s);
#endif

#endif
		unlock_mks();

} 

int mozna_czytac(int fd)
{
  fd_set secik;
  struct timeval timeout;
  int retval;

	timeout.tv_sec=1;
	timeout.tv_usec=0;
	FD_ZERO(&secik);
	FD_SET(fd,&secik);
	retval = select(fd+1, &secik,	0, 0, &timeout);
	FD_ZERO(&secik);
	return retval;
}

void kto_zacz(addr_info)
        struct sockaddr_in *addr_info;
{
        int             sockfd;
        struct sockaddr_in gdzie_addr;
        char        addr[128],
                    query[MAXBUFF];
        struct stat buf;
        FILE*   q;

        // ident ?
    syslog(LOG_INFO, "%ld %s", ntohs(addr_info->sin_port),
    //                ntohl(addr_info->sin_addr.s_addr),
                    inet_ntoa(addr_info->sin_addr.s_addr));

//        get ident data:
        gdzie_addr.sin_family = AF_INET;
        gdzie_addr.sin_addr.s_addr = addr_info->sin_addr.s_addr; // inet_addr(addr);
        gdzie_addr.sin_port = htons(113);

	if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
		syslog(LOG_INFO, "can't open ident socket to %s", inet_ntoa(addr_info->sin_addr.s_addr));
	else
		if (connect (sockfd, (struct sockaddr *) &gdzie_addr,
			sizeof(gdzie_addr)) < 0) {
		syslog(LOG_INFO, "can't connect ident socket to %s", inet_ntoa(addr_info->sin_addr.s_addr));
	        } else {
        q=fdopen(sockfd,"r+");
        fprintf(q,"%d,%d\n", ntohs(addr_info->sin_port), SERV_PORT);
        bzero(query,MAXBUFF);
        fgets(query,MAXBUFF,q);

        syslog(LOG_INFO, "%ld %s ident data follows: %s", ntohs(addr_info->sin_port),
            inet_ntoa(addr_info->sin_addr.s_addr), query);
        
	}
}
        
void	obsluz_klienta(newsockfd,mks_in, mks_out, mks_err)
{
		FILE *m_fifo, *f_tmp;
		char s[MAXBUFF], f[10];
		int i;

		alarm(CLIENT_TIMEOUT);
		signal(SIGALRM, (sig_t) on_child_timeout);
        
		pobierz_dane(newsockfd,&m_fifo,&i);

    shutdown(newsockfd,SHUT_RD);
		zrob_chwilowy(&f_tmp,&f);

		if (i>0) {
				kopiuj_plik_FILE(i, m_fifo, f_tmp);
				fclose(f_tmp);
				skanuj(mks_in,mks_out,mks_err,f,m_fifo);
#ifdef DEBUG_LOGCLIENTS
				syslog(LOG_INFO, "%s Length: %d, childs: %d", f, i, child_count);
#endif
		} else {
                i=-i;
                syslog(LOG_INFO,"Status Information %d ",i);
                if ((i&1)==1) fprintf(m_fifo,"%s",mks_version);
                if ((i&2)==2) statystyki_file(m_fifo);
                if ((i&4)==4) fprintf(m_fifo,"%s",mks_bversion);
        }

		unlink(f);

		fflush(m_fifo);
		fclose(m_fifo);
        shutdown(newsockfd,SHUT_RDWR);
		close(newsockfd);
		//koniec klienta.
        close_all_fds();
        exit(0);
}

int main(char **argv,int argc)
{
		int rura_a, rura_b, rura_c, fd;
		int fd_a[2], fd_b[2], fd_c[2];
		FILE *mks_in, *mks_out, *mks_err;
		struct sockaddr_in cli_addr, serv_addr;
		int sockfd, newsockfd, clilen, tmp;
		pid_t	child_pid,synek;
		extern child_count;

	init_logs();
	init_socket(&sockfd);
    umask(0077);
    
	if (unlink(MKS_LOCK_FILE)==0) syslog(LOG_INFO, "udalo sie usunac MKS_LOCK");
    
	odepnij_tty();
    zrob_rury(&rura_a, &rura_b, &rura_c, fd_a, fd_b, fd_c, &mks_in, &mks_out, &mks_err);
    ignoruj();

    // ustawiamy flag CLOSE-ON-EXEC (bez tego mks32 nie zakoczy si po
    // zamkiciu deskryptorw w tatuiu
    fcntl(fd_a[0], F_SETFD, FD_CLOEXEC);
    fcntl(fd_a[1], F_SETFD, FD_CLOEXEC);
    fcntl(rura_a, F_SETFD, FD_CLOEXEC);
    fcntl(rura_b, F_SETFD, FD_CLOEXEC);
    fcntl(rura_c, F_SETFD, FD_CLOEXEC);

    if ((synek = fork()) == -1) {
		err("There is no fork !");
    } else {
				
		if (synek == 0) {
	    	/*
		     * synek  proces mks32
		     */
			run_mks(fd_a[0],fd_b[1],fd_c[1]);
	} else {
	    /*
	     * tato 
	     */

    tmp=init_mks(mks_in, mks_out, mks_err);
		fflush(mks_out);
		fflush(mks_in);
		fflush(mks_err);

		if (tmp==0) {


	        if ((tmp=fork())<0) 
				err("There is no knife!");
			else if (tmp==0)
			{
				clilen = sizeof(cli_addr);
				signal(SIGCHLD, (sig_t) on_child_exit);
#ifdef DEBUG_STATS
				statystyki();
#endif
				while (1) {
				    ignoruj();
						signal(SIGTERM, (sig_t) on_signal);	// sig_t is glibc specific ???
						if ((child_count<MAX_CLIENTS) && (mozna_czytac(sockfd)>0) && (exit_flag==0) && (reload_flag==0)) {
								if ((newsockfd = accept(sockfd, (struct sockaddr*) &cli_addr, &clilen)) < 0)
										err ( "accept error");
								if ((child_pid = fork()) < 0) err("there is no spoon!\n");
								else 
										// Nowy klient
										if (child_pid == 0) {	
												close(sockfd);
#ifdef USE_IDENT
                                                kto_zacz(&cli_addr);
#endif                                                
												obsluz_klienta(newsockfd, mks_in, mks_out, mks_err);
										}
										else if (child_pid > 0) {
												// tato
												child_count++;
												close(newsockfd);
										};	// if child_pid>0
			};  // if child_count
			if (exit_flag==1) {
					signal(SIGTERM, SIG_IGN);
					// Czekamy a synkowie si zakocz (mog jeszcze co skanowa)
					// maj na to CLIENT_TIMEOUT sekund (domylnie 10)
					while (child_count>0)
						while (waitpid(-1, NULL,WNOHANG)>0) { child_count--; child_served++; }
					// zamknicie synka:
					lock_mks();
					end_mks(fd_a, fd_b, fd_c, mks_in, mks_out, mks_err);
					unlock_mks();
					if (synek > 0) {
						kill(synek, SIGTERM);
						waitpid(synek, NULL, 0);
					}
					// zapiszemy si w syslogu
					statystyki();
					syslog(LOG_NOTICE,"Exit");
					unlink(MKS_LOCK_FILE);
					close_all_fds();
					exit(0);
				} // if (exit_flag
			if (reload_flag==1)
					{
							syslog(LOG_NOTICE,"Reload.");
							lock_mks();
							end_mks(fd_a, fd_b, fd_c, mks_in, mks_out, mks_err);
							unlock_mks();
							if (synek > 0) {
									kill(synek, SIGTERM);
									waitpid(synek, NULL, 0);
							}
					   fclose(mks_out);
					   fclose(mks_in);
					   fclose(mks_err);
    zrob_rury(&rura_a, &rura_b, &rura_c, fd_a, fd_b, fd_c, &mks_in, &mks_out, &mks_err);
		    fcntl(fd_a[0], F_SETFD, FD_CLOEXEC);
				    fcntl(fd_a[1], F_SETFD, FD_CLOEXEC);
						    fcntl(rura_a, F_SETFD, FD_CLOEXEC);
								    fcntl(rura_b, F_SETFD, FD_CLOEXEC);
										    fcntl(rura_c, F_SETFD, FD_CLOEXEC);

						syslog(LOG_NOTICE,"Reload.");
								

							if ((synek = fork()) == -1) 
						    err("There is no fork !"); 
								if (synek==0) 
								{
								run_mks(fd_a[0],fd_b[1],fd_c[1]);
								exit(-100);
							}
						syslog(LOG_NOTICE,"Reload..");
							tmp=init_mks(mks_in, mks_out, mks_err);
							if(tmp!=0) exit(tmp);
					    reload_flag=0;
					} // if (reload_flag
			} // while
		}; //else tato
		};
		if (tmp>0) {tmp=0;}
	 }; //if tmp
	}
	close_all_fds();
    return tmp;
}
// $Id: mks_d.c,v 1.2 2003/08/17 15:25:33 hunter Exp $
// vi:tw=78:ts=2:sw=4
